/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2001 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 ****************************************************************************/

#include "config.h"
#include <xpl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <mdb.h>

#define STR unsigned char
#define CONSTSTR const unsigned char
#define MAX_LIST 64

#define EXT_USAGE_MSG "Usage: mdbtool [options] <command>\n\n"\
"Options:\n"\
"-h, --help      Display this message\n"\
"-p, --password  Password for authentication\n"\
"-u, --username  Username for authentication\n\n"\
"Commands:\n"\
"getattrs <object> <attrname>\n"\
"getdnattrs <object> <attrname>\n"\
"addattr <object> <attrname> <value>\n"\
"setattr <object> <attrname> <value>\n"\
"addobject <object> <class> [attr1=value1 attr2=...]\n"\
"delobject <object>\n"\
"setpassword <object> <password|-f pwfile>\n"\
"verifypassword <object> <password>\n"\
"isobject <object>\n"\
"defattr <attrname> <asn1> [opt1=val1 opt2=...]\n"\
"defclass <classname> <asn1> [opt1=val1 opt2=...]\n"\
"defclassattr <classname> <attrname>\n"\
"delattrdef <attrname>\n"\
"delclassdef <classname>\n\n"

struct dict {
    void *next;
    STR *key;
    STR *value;
};

void DumpAttr(MDBHandle *handle, MDBValueStruct *vs, CONSTSTR *object);
void GetAttr(MDBHandle *h, CONSTSTR *object, CONSTSTR *attr);
void GetDnAttr(MDBHandle *h, CONSTSTR *object, CONSTSTR *attr);
BOOL AddAttr(MDBHandle *h, CONSTSTR *object, CONSTSTR *attr, CONSTSTR *value);
BOOL SetAttr(MDBHandle *h, CONSTSTR *object, CONSTSTR *attr, CONSTSTR *value);
BOOL AddObject(MDBHandle *h, CONSTSTR *object, CONSTSTR *class, struct dict *attrs); 
BOOL IsObject(MDBHandle *h, CONSTSTR *object);
BOOL SetPassword(MDBHandle *h, CONSTSTR *object, CONSTSTR *password); 
BOOL VerifyPassword(MDBHandle *h, CONSTSTR *object, CONSTSTR *password);
BOOL DelObject(MDBHandle *h, CONSTSTR *object, BOOL recursive);
BOOL DefAttr(MDBHandle *h, CONSTSTR *attrname, CONSTSTR *asn1, unsigned long type, BOOL single, BOOL sync, BOOL public);
BOOL DefClass(MDBHandle *h, CONSTSTR *classname, CONSTSTR *asn1, BOOL container, CONSTSTR *super, CONSTSTR *name, char **containers, char **requires, char **allows);
BOOL DefClassAttr(MDBHandle *h, CONSTSTR *classname, CONSTSTR *attrname);
BOOL DelAttrDef(MDBHandle *h, CONSTSTR *attrname);
BOOL DelClassDef(MDBHandle *h, CONSTSTR *classname);
struct dict *GetDict(char *list[], int items);
void FreeDict(struct dict *d);

int main (int argc, char *argv[]) {
    STR *command;
    STR *object;
    STR *auth_username = NULL;
    STR *auth_password = NULL;
    int next_arg = 1;
    STR *attrname;
    STR *value;
    STR *type;
    struct dict *attrs;
    STR *rec_arg;
    BOOL recursive;
    STR *arg_ptr;
    STR *password;
    FILE *pwfile;
    int next_char;
    int char_count;
    unsigned long syn_type;
    BOOL single;
    BOOL sync;
    BOOL public;

    /* parse optional attributes */    
    while (next_arg < argc && argv[next_arg][0] == '-') {
        if (next_arg + 1 < argc) {
            /* authorization credentials */
            if (!strcmp(argv[next_arg], "-u") || !strcmp(argv[next_arg], "--username")) {
                auth_username = argv[++next_arg];
            } else if (!strcmp(argv[next_arg], "-p") || !strcmp(argv[next_arg], "--password")) {
                auth_password = argv[++next_arg];
            } else if (!strcmp(argv[next_arg], "-h") || !strcmp(argv[next_arg], "--help")) {
                printf(EXT_USAGE_MSG);
            }

            next_arg++;
        } else {
            printf("mdbtool: Unrecognized option: %s\n\n", argv[next_arg]);
            printf(EXT_USAGE_MSG);
            return 0;
        }
    }
    
    command = (STR*)argv[next_arg++];
    object = (STR*)argv[next_arg++];
    
    if ((command == NULL)||(object == NULL)) {
        /*printf("Usage: mdbtool <command> <object> ...\n");*/
        printf(EXT_USAGE_MSG);
        return 0;
    }
    
    if (!MDBInit()) {
        printf("ERROR: Could not initialise MDB\n");
        return 1;
    }
    
    MDBHandle *h = MDBAuthenticate((STR*)"Hula", auth_username, auth_password);

    if (h == NULL) {
        printf("ERROR: Could not authenticate to MDB (%s)\n", auth_username);
        return 2;
    }
    
    if (strcmp((const char*)command, "getattrs") == 0) {
        attrname = (STR*)argv[next_arg++];

        if (attrname == NULL) {
            printf("Usage: mdbtool getattrs <object> <attrname>\n");
            return 0;
        }

        GetAttr(h, object, attrname);
    } else if (strcmp((const char*)command, "getdnattrs") == 0) {
        attrname = (STR*)argv[next_arg++];

        if (attrname == NULL) {
            printf("Usage: mdbtool getdnattrs <object> <attrname>\n");
            return 0;
        }

        GetDnAttr(h, object, attrname);
    } else if (strcmp((const char*)command, "addattr") == 0) {
        attrname = (STR*)argv[next_arg++];
        value = (STR*)argv[next_arg++];
        
        if ((attrname == NULL) || (value == NULL)) {
            printf("Usage: mdbtool addattr <object> <attrname> <value>\n");
            return 0;
        }

        if (!AddAttr(h, object, attrname, value)) {
            printf("ERROR: Could not add attribute\n");
            return 1;
        }
    } else if (strcmp((const char*)command, "setattr") == 0) {
        attrname = (STR*)argv[next_arg++];
        value = (STR*)argv[next_arg++];
        
        if ((attrname == NULL) || (value == NULL)) {
            printf("Usage: mdbtool setattr <object> <attrname> <value>\n");
            return 0;
        }

        if (!SetAttr(h, object, attrname, value)) {
            printf("ERROR: Could not set attribute\n");
            return 1;
        }
    } else if (strcmp((const char*)command, "addobject") == 0) {
        type = (STR*)argv[next_arg++];

        if (type == NULL) {
            printf("Usage: mdbtool addobject <object> <class> [attr1=value1 attr2=...]\n");
            return 0;
        }

        attrs = GetDict(&argv[next_arg], argc-next_arg);

        if (AddObject(h, object, type, attrs)) {
            printf("ERROR: Could not add object\n");
            return 1;
        }

        FreeDict(attrs);
    } else if (strcmp((const char*)command, "delobject") == 0) {
        rec_arg = (STR*)argv[next_arg++];
        recursive = 0;

        if (rec_arg == NULL || !strcasecmp(rec_arg, "false")) {
            recursive = 0;
        } else if (!strcasecmp(rec_arg, "true")) {
            recursive = 1;
        }

        if (DelObject(h, object, recursive)) {
            printf("ERROR: Could not delete object\n");
            return 1;
        }
    } else if (strcmp((const char*)command, "setpassword") == 0) {
        next_char = 0;
        char_count = 0;
        arg_ptr = (STR*)argv[next_arg++];
        
        if (!strcmp(arg_ptr, "-f") && next_arg <= argc) {
            arg_ptr = (STR*)argv[next_arg++];
            pwfile = fopen(arg_ptr, "r");
            password = malloc(128);

            while ((next_char = fgetc(pwfile)) != EOF && char_count < 128) {
                password[char_count++] = next_char;
            }

            if (next_char != EOF) {
                printf("ERROR: Password file too long (128 bytes max).\n");
                free(password);
                return 1;
            }

            fclose(pwfile);
        } else {
            password = arg_ptr;
            arg_ptr = NULL;
        }
        
        if (SetPassword(h, object, password)) {
            printf("ERROR: Could not change the password.\n");
            return 1;
        }

        if (arg_ptr) {
            free(password);
        }
    } else if (strcmp((const char *)command, "verifypassword") == 0) {
        password = (STR*)argv[next_arg++];

        if (VerifyPassword(h, object, password)) {
            printf("ERROR: Could not verify password.\n");
            return 1;
        }
    } else if (strcmp((const char *)command, "isobject") == 0) {
        IsObject(h, object);

    } else if (strcmp((const char *)command, "defattr") == 0) {
        STR *asn1 = (STR*)argv[next_arg++];    
        syn_type = MDB_ATTR_SYN_STRING;
        single = 0;
        sync = 0;
        public = 0;
        attrname = object;

        if (attrname == NULL || asn1 == NULL) {
            printf("Usage: mdbtool defattr <attrname> <asn1> [opt1=val1 opt2=...]\n");
            return 0;
        }
        
        attrs = GetDict(&argv[next_arg], argc - next_arg);
        
        while (attrs != NULL) {
            if ((attrs->key != NULL) && (attrs->value != NULL)) {
                if (!strcmp(attrs->key, "type")) {
                    if (!strcasecmp(attrs->value, "string")) {
                        syn_type = MDB_ATTR_SYN_STRING;
                    } else if (!strcasecmp(attrs->value, "dn")) {
                        syn_type = MDB_ATTR_SYN_DIST_NAME;
                    } else {
                        printf("mdbtool: Option 'type' invalid (must be 'string' or 'dn').\n");
                        return 0;
                    }
                } else if (!strcmp(attrs->key, "singleval")) {
                    if (!strcasecmp(attrs->value, "true")) {
                        single = 1;
                    } else if (!strcasecmp(attrs->value, "false")) {
                        single = 0;
                    } else {
                        printf("mdbtool: Option 'singleval' invalid (must be 'true' or 'false').\n");
                        return 0;
                    }
                } else if (!strcmp(attrs->key, "immsync")) {
                    if (!strcasecmp(attrs->value, "true")) {
                        sync = 1;
                    } else if (!strcasecmp(attrs->value, "false")) {
                        sync = 0;
                    } else {
                        printf("mdbtool: Option 'immsync' invalid (must be 'true' or 'false').\n");
                        return 0;
                    }
                } else if (!strcmp(attrs->key, "public")) {
                    if (!strcasecmp(attrs->value, "true")) {
                        public = 1;
                    } else if (!strcasecmp(attrs->value, "false")) {
                        public = 0;
                    } else {
                        printf("mdbtool: Option 'public' invalid (must be 'true' or 'false').\n");
                        return 0;
                    }
                } else {
                    printf("mdbtool: Unsupported option '%s'\n", attrs->key);
                    return 0;
                }
            }
    
            attrs = attrs->next;
        }

        if (DefAttr(h, attrname, asn1, syn_type, single, sync, public)) {
            printf("ERROR: Could not define attribute\n");
            return 1;
        }

        FreeDict(attrs);
    } else if (strcmp((const char *)command, "defclass") == 0) {
        BOOL container = FALSE;
        char *superclass = NULL;
        char *naming = NULL;
        char *containers[MAX_LIST];
        char *requires[MAX_LIST];
        char *allows[MAX_LIST];
        int next_containers = 0;
        int next_requires = 0;
        int next_allows = 0;
        char *nptr;
        STR *classname = object;
        STR *asn1 = (STR*)argv[next_arg++];

        if (classname == NULL || asn1 == NULL) {
            printf("Usage: mdbtool defclass <classname> <asn1> [opt1=val1 opt2=...]\n");
            return 0;
        }
        
        attrs = GetDict(&argv[next_arg], argc - next_arg);
        
        while (attrs != NULL) {
            if ((attrs->key != NULL) && (attrs->value != NULL)) {
                if (!strcmp(attrs->key, "container")) {
                    if (!strcasecmp(attrs->value, "true")) {
                        container = TRUE;
                    } else if (!strcasecmp(attrs->value, "false")) {
                        container = FALSE;
                    } else {
                        printf("mdbtool: Option 'container' invalid (must be 'true' or 'false').\n");
                        return 0;
                    }
                } else if (!strcmp(attrs->key, "superclass")) {
                    superclass = attrs->value;
                } else if (!strcmp(attrs->key, "containers")) {
                    nptr = attrs->value;
                    containers[next_containers++] = nptr;

                    while (next_containers < MAX_LIST && (nptr = strchr(nptr, ',')) != '\0') {
                        *nptr++ = '\0';
                        containers[next_containers++] = nptr;
                    }
                } else if (!strcmp(attrs->key, "naming")) {
                    naming = attrs->value;
                } else if (!strcmp(attrs->key, "requires")) {
                    nptr = attrs->value;
                    requires[next_requires++] = nptr;

                    while (next_requires < MAX_LIST && (nptr = strchr(nptr, ',')) != '\0') {
                        *nptr++ = '\0';
                        requires[next_requires++] = nptr;
                    }
                } else if (!strcmp(attrs->key, "allows")) {
                    nptr = attrs->value;
                    allows[next_allows++] = nptr;

                    while (next_allows < MAX_LIST && (nptr = strchr(nptr, ',')) != '\0') {
                        *nptr++ = '\0';
                        allows[next_allows++] = nptr;
                    }
                }
            }

            attrs = attrs->next;
        }

        containers[next_containers] = NULL;
        requires[next_requires] = NULL;
        allows[next_allows] = NULL;

    	if (!superclass) {
            printf("mdbtool: Option 'superclass' is required.\n");
	       return 0;
        }

        if (DefClass(h, classname, asn1, container, superclass, naming, containers, requires, allows)) {
            printf("ERROR: Could not define class\n");
            return 1;
        }

        FreeDict(attrs);
    } else if (strcmp((const char *)command, "defclassattr") == 0) {
        attrname = (STR *)argv[next_arg++];

    	if (!attrname) {
            printf("mdbtool: attribute not specified\n");
            return 0;
	    }
	
        if (DefClassAttr(h, object, attrname)) {
            printf("ERROR: Could not add attribute to class\n");
            return 1;
        }
    } else if (strcmp((const char*)command, "delattrdef") == 0) {
        if (DelAttrDef(h, object)) {
            printf("ERROR: Could not remove attribute definition\n");
            return 1;
        }
    } else if (strcmp((const char*)command, "delclassdef") == 0) {
        if (DelClassDef(h, object)) {
            printf("ERROR: Could not remove class definition\n");
            return 1;
        }
    } else {
        printf("mdbtool: Unsupported command '%s'.\n", command);
    }    
    
    return 0;
}

void GetAttr(MDBHandle *h, CONSTSTR *object, CONSTSTR *attr) {
    MDBValueStruct *vs;
    long i;
    vs = MDBCreateValueStruct(h, NULL);
    long l = MDBRead(object, attr, vs);
    
    for (i=0; i<l; i++) {
        printf ("%s\n", vs->Value[i]);
    }
    
    MDBDestroyValueStruct(vs);
}

void GetDnAttr(MDBHandle *h, CONSTSTR *object, CONSTSTR *attr) {
    MDBValueStruct *vs;
    long i;
    vs = MDBCreateValueStruct(h, NULL);
    long l = MDBReadDN(object, attr, vs);
    
    for (i = 0; i < l; i++) {
        printf ("%s\n", vs->Value[i]);
    }
    
    MDBDestroyValueStruct(vs);
}

BOOL AddAttr(MDBHandle *h, CONSTSTR *object, CONSTSTR *attr, CONSTSTR *value) {
    MDBValueStruct *vs;
    BOOL result;

    vs = MDBCreateValueStruct(h, NULL);
    
    if (strchr(value, '\\') != '\0') {
        result = MDBAddDN(object, attr, value, vs);
    } else {
        result = MDBAdd(object, attr, value, vs);
    }

    MDBDestroyValueStruct(vs);

    return result;
}

BOOL SetAttr(MDBHandle *h, CONSTSTR *object, CONSTSTR *attr, CONSTSTR *value) {
    MDBValueStruct *vs;
    BOOL result;
    
    vs = MDBCreateValueStruct(h, NULL);
    MDBAddValue(value, vs);

    if (strchr(value, '\\') != '\0') {
        result = MDBWriteDN(object, attr, vs);
    } else {
        result = MDBWrite(object, attr, vs);
    }
    
    MDBDestroyValueStruct(vs);
    
    return result;
}

BOOL AddObject(MDBHandle *h, CONSTSTR *object, CONSTSTR *class, struct dict *attrs) {
    MDBValueStruct *attr, *value, *vs;
    BOOL result;
    
    attr = MDBCreateValueStruct(h, NULL);
    value = MDBCreateValueStruct(h, NULL);
    vs = MDBCreateValueStruct(h, NULL);

    while (attrs != NULL) {
        if ((attrs->key != NULL) && (attrs->value != NULL)) {
            /* not sure if this will always be accurate */
            if (strchr(attrs->value, '\\') != '\0') {
                MDBAddDNAttribute(attrs->key, attrs->value, attr, value);
            } else {
                MDBAddStringAttribute(attrs->key, attrs->value, attr, value);
            }
            /*
            MDBAddValue(attrs->key, attr);
            MDBAddValue(attrs->value, value);
	    */
        }

        attrs = attrs->next;
    }
    
    result = MDBCreateObject(object, class, attr, value, vs);
    
    MDBDestroyValueStruct(vs);
    MDBDestroyValueStruct(value);
    MDBDestroyValueStruct(attr);
    return (!result);
}

BOOL SetPassword(MDBHandle *h, CONSTSTR *object, CONSTSTR *password) {
    MDBValueStruct *vs;
    BOOL result;
    
    vs = MDBCreateValueStruct(h, NULL);
    
    result = MDBChangePasswordEx(object, NULL, password, vs);
    
    MDBDestroyValueStruct(vs);
    return (!result);
}

BOOL VerifyPassword(MDBHandle *h, CONSTSTR *object, CONSTSTR *password) {
    MDBValueStruct *vs;
    BOOL result;
    
    vs = MDBCreateValueStruct(h, NULL);
    result = MDBVerifyPassword(object, password, vs);
    
    MDBDestroyValueStruct(vs);
    return (!result);
}

BOOL IsObject(MDBHandle *h, CONSTSTR *object) {
    MDBValueStruct *vs;
    BOOL result;

    vs = MDBCreateValueStruct(h, NULL);
    if ((result = MDBIsObject(object, vs))) {
        printf("%s\n", object);
    }
    
    MDBDestroyValueStruct(vs);
    return (result);
}

BOOL DelObject(MDBHandle *h, CONSTSTR *object, BOOL recursive) {
    MDBValueStruct *vs;
    BOOL result;

    vs = MDBCreateValueStruct(h, NULL);
    result = MDBDeleteObject(object, recursive, vs);
    
    MDBDestroyValueStruct(vs);
    return (!result);
}

BOOL DefAttr(MDBHandle *h, CONSTSTR *attrname, CONSTSTR *asn1, unsigned long type, BOOL single, BOOL sync, BOOL public) {
    MDBValueStruct *vs;
    BOOL result;
    
    vs = MDBCreateValueStruct(h, NULL);
    result = MDBDefineAttribute(attrname, asn1, type, single, sync, public, vs);

    MDBDestroyValueStruct(vs);
    return (!result);
}

BOOL DefClass(MDBHandle *h, CONSTSTR *classname, CONSTSTR *asn1, BOOL container, CONSTSTR *super, CONSTSTR *name, char **containers, char **requires, char **allows) {
    MDBValueStruct *vs;
    MDBValueStruct *superclass, *containment, *naming, *mandatory, *optional;
    BOOL result;
    int i;

    vs = MDBCreateValueStruct(h, NULL);
    superclass = MDBCreateValueStruct(h, NULL);
    containment = MDBCreateValueStruct(h, NULL);
    naming = MDBCreateValueStruct(h, NULL);
    mandatory = MDBCreateValueStruct(h, NULL);
    optional = MDBCreateValueStruct(h, NULL);
    
    if (super) {
       MDBAddValue(super, superclass);
    }

    if (name) {
        MDBAddValue(name, naming);
    }

    i = 0;
    while (i < MAX_LIST && containers[i]) {
        MDBAddValue(containers[i++], containment);
    }

    i = 0;
    while (i < MAX_LIST && requires[i]) {
        MDBAddValue(requires[i++], mandatory);
    }

    i = 0;
    while (i < MAX_LIST && allows[i]) {
        MDBAddValue(allows[i++], optional);
    }
    
    result = MDBDefineClass(classname, asn1, container, superclass, containment, naming, mandatory, optional, vs);

    MDBDestroyValueStruct(superclass);
    MDBDestroyValueStruct(containment);
    MDBDestroyValueStruct(naming);
    MDBDestroyValueStruct(mandatory);
    MDBDestroyValueStruct(optional);

    return (!result);
}

BOOL DefClassAttr(MDBHandle *h, CONSTSTR *classname, CONSTSTR *attrname) {
    MDBValueStruct *vs;
    BOOL result;

    vs = MDBCreateValueStruct(h, NULL);
    result = MDBAddAttribute(attrname, classname, vs);
    MDBDestroyValueStruct(vs);

    return (!result);
}

BOOL DelAttrDef(MDBHandle *h, CONSTSTR *attrname) {
    MDBValueStruct *vs;
    BOOL result;

    vs = MDBCreateValueStruct(h, NULL);
    result = MDBUndefineAttribute(attrname, vs);
    MDBDestroyValueStruct(vs);

    return (!result);
}

BOOL DelClassDef(MDBHandle *h, CONSTSTR *classname) {
    MDBValueStruct *vs;
    BOOL result;

    vs = MDBCreateValueStruct(h, NULL);
    result = MDBUndefineClass(classname, vs);
    MDBDestroyValueStruct(vs);

    return (!result);
}

struct dict *GetDict(char *list[], int items) {
    struct dict *d;
    char *item;
    
    if (items == 0) return NULL;
    item = list[0];
    
    d = malloc(sizeof(struct dict));
    d->key = (STR *)item;
    d->value = (STR *)strchr(item, '=');
    *d->value = 0;
    d->value++;
    d->next = GetDict(&list[1], items - 1);
    
    return d;
}

void FreeDict(struct dict *d) {
    struct dict *tmp;
    
    while (d != NULL) {    
        tmp = d;
        d = d->next;
        free(tmp);
    }
}

/* broken vv */

void DumpAttr(MDBHandle *handle, MDBValueStruct *vs, CONSTSTR *object) {
    /* MDBEnumStruct *e; */
    MDBValueStruct *v;
    CONSTSTR *attr = NULL;
    
    v = MDBCreateValueStruct(handle, object);
    /*attr = MDBEnumerateAttributesEx(object, vs); FIXME: API */
    while(attr != NULL) {
        long l = MDBRead(object, attr, v);
        if (l != 0) {
            long i;
            for (i=0; i<l; i++) {
                printf("\t%s: %s\n", attr, v->Value[i]);
            }
        }
        /*attr = MDBEnumerateAttributesEx(object, vs);*/
    }
    
    MDBDestroyValueStruct(v);
}
