using System;
using System.IO;
using System.Collections;
using System.Runtime.InteropServices;
using Lucene.Net.Documents;

namespace Hula.ICal {

internal class ICalRecurrenceEnumerable : IEnumerable {
    
    internal class EmptyEnumerator : IEnumerator {
	public object Current {
	    get { throw new InvalidOperationException(); }
	}	
	public bool MoveNext() {
	    return false;
	}
	public void Reset() {
	} 
    }

    internal class ICalRecurrenceEnumerator : IEnumerator, IDisposable {
	ICalRecurrence recur;
	ICalTime start;
	ICalTime thisTime;
	IntPtr iter = IntPtr.Zero;
	bool done;
	
	[DllImport(ICalComponent.icalLib)]
	private static extern IntPtr icalrecur_iterator_new(ICalRecurrence recur, ICalTime start);

	[DllImport(ICalComponent.icalLib)]
	private static extern ICalTime icalrecur_iterator_next(IntPtr iter);

	[DllImport(ICalComponent.icalLib)]
	private static extern void icalrecur_iterator_free(IntPtr iter); 
	
	internal ICalRecurrenceEnumerator(ICalRecurrence recur, ICalTime start) {
	    thisTime = start;
	    this.start = start;
	    this.recur = recur;
	    Reset();
	}

	public object Current {
	    get {
		if (done == true) {
		    throw new InvalidOperationException();
		}
		return thisTime; 
	    }
	}

	public bool MoveNext() {
	    if (done == true) {
		return false;
	    }

	    thisTime = icalrecur_iterator_next(iter);
	    if (thisTime.IsNull) {
		icalrecur_iterator_free(iter);
		iter = IntPtr.Zero;
		done = true;
		return false;
	    }

	    return true;
	}

	public void Reset() {
	    if (iter != IntPtr.Zero) {
		icalrecur_iterator_free(iter);
		iter = IntPtr.Zero;
	    }
	    iter = icalrecur_iterator_new(recur, start);
	    done = false;
	}

	public void Dispose() {
	    if (iter != IntPtr.Zero) {
		icalrecur_iterator_free(iter);
		iter = IntPtr.Zero;
	    }
	    GC.SuppressFinalize(this);
	}

	~ICalRecurrenceEnumerator() {
	    Dispose();
	}
    }

    ICalComponent comp;
    ICalProperty ruleProp;
    internal ICalRecurrenceEnumerable (ICalComponent c, ICalProperty ruleProp) {
	comp = c;
	this.ruleProp = ruleProp;
    }

    public IEnumerator GetEnumerator() {
	if (ruleProp != null) {
	    ICalRecurrence rule = ruleProp.RecurrenceRule;
	    return new ICalRecurrenceEnumerator(rule, comp.StartTime);
	} else {
	    return new EmptyEnumerator();
	}
    }
}

[StructLayout(LayoutKind.Sequential)]
public struct ICalRecurrence {
    public enum Frequency {
	Secondly,
	Minutely,
	Hourly,
	Daily,
	Weekly,
	Monthly,
	Yearly,
	None
    };

    public Frequency frequence;

    public ICalTime until;

    public int count;

    public short interval;

    public enum Weekday {
	None,
	Sunday,
	Monday,
	Tuesday,
	Wednesday,
	Thursday,
	Friday,
	Saturday
    };

    public Weekday weekStart;

    [MarshalAs(UnmanagedType.ByValArray, SizeConst=61)] public short[] bySecond;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst=61)] public short[] byMinute;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst=25)] public short[] byHour;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst=364)] public short[] byDay;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst=32)] public short[] byMonthDay;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst=367)] public short[] byYearDay;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst=54)] public short[] byWeekNumber;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst=13)] public short[] byMonth;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst=367)] public short[] bySetPos;
};
}
