using System;
using System.Collections;
using System.IO;
using System.Text;
using System.Threading;

using Hula.Nmap;
using Indexing = Hula.Indexing;
using Hula.Cal;
using Hula.Indexing;
using Hula.Store;
using Hula.Util;

using Lucene.Net;
using Lucene.Net.Analysis;
using Lucene.Net.Analysis.Standard;
using Lucene.Net.Documents;
using Lucene.Net.Index;
using Lucene.Net.QueryParsers;
using Lucene.Net.Search;
using Lucene.Net.Store;

using log4net;

namespace Hula.Indexer
{
public class IndexClient
{
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    private int userCount;

    private HulaThreadPool pool = null;

    private static int INDEX_VERSION = 1;

    internal struct IndexArgs
    {
        public IndexArgs(string user, string[] guids, HulaThreadPriority priority)
        {
            this.user = user;
            this.guids = guids;
            this.priority = priority;
        }

        public string user;
        public string[] guids;
        public HulaThreadPriority priority;
    }

    public IndexClient()
    {
        pool = new HulaThreadPool(5, 30);
    }

    public void Close()
    {
    }

    private LuceneIndex GetIndex (string user)
    {
        LuceneIndex idx = IndexFactory.GetIndex(user);

        return idx;
    }

    public void IndexNMAPStore()
    {
        log.Debug("indexing an NMAP store");
        userCount = 0;

        try {
            using (NmapClient nc = new NmapClient()) {
                IList users = nc.Open.ULIST();

                foreach (string user in users)
                {
                    Interlocked.Increment (ref userCount);
                    QueueIndexUser(user, "NMAP", null, HulaThreadPriority.Bulk);
                }
            }
        } catch (System.Net.Sockets.SocketException e) {
            log.Warn("Could not connect to local NMAP store - disabling bootstrap index");
        }
    }

    public void QueueIndexUser (string user, string store,
                                string[] guids, HulaThreadPriority priority)
    {
        IndexArgs args = new IndexArgs(user, guids, priority);
        pool.QueueWorkToDo(new WaitCallback(IndexNMAPUser), args, priority);
    }

    public void IndexUser (string user, string store, string[] guids,
                           HulaThreadPriority priority)
    {
        IndexArgs args = new IndexArgs(user, guids, priority);

        if (store.Equals("NMAP") 
            || store.Equals("mail") 
            || store.Equals("calendar") 
            || store.Equals("addressbook")) {
            IndexNMAPUser(args);
        } else {
            log.ErrorFormat("Unrecognized store: {0}", store);
        }
    }

    private void IndexNMAPUser (object data)
    {
        IndexArgs args = (IndexArgs)data;

        /* Really bad hack */
        string user = String.Intern(args.user);
        lock (user) {
            log.DebugFormat("indexing an NMAP store for user: {0}", user);
            
            NmapClient client = new NmapClient (user);
            client.Auth.USER (user);
            
            LuceneIndex idx = GetIndex (user);

            try
            {
                using(idx) {
                    IndexCurrentUser (args, client, idx);
                }
            } catch (Exception e) {
                log.ErrorFormat("Exception indexing user {0}: {1}", user, e);
            }
            
            client.Close ();
        }
        Interlocked.Decrement (ref userCount);
    }

    private void IndexCurrentUser(IndexArgs args, NmapClient client, LuceneIndex idx)
    {
        string user = args.user;
        string[] guids = args.guids;

        log.DebugFormat("indexing user: {0}", user);

        Indexing.Indexer ab = new VCardIndexer();
        Indexing.Indexer cal = new CalIndexer();
        Indexing.Indexer folder = new Indexing.FolderIndexer();
        Indexing.Indexer mail = new MailIndexer();

        IList documents = null;

        // Find documents purged since the last run
        documents = client.Document.DLSPG();

        if (documents.Count > 0) {
            // Remove them from the index
            foreach (string guid in documents) {
                log.DebugFormat("Removing purged document from the index: {0}", guid);
                idx.Reader.Delete(new Term("guid", guid));
            }

            // Notify NMAP that we've removed them
            client.Document.DRMPG(documents);
        }

        if (guids == null) {
            documents = client.Document.DUNFO();
        } else {
            documents = new ArrayList();

            IList infos = client.Document.DINFO(guids);
            foreach (DocumentInfo info in infos) {
                if (info.indexver != IndexClient.INDEX_VERSION) {
                    documents.Add(info);
                } else {
                    log.DebugFormat("Skipping document {0} (already indexed)", info.guid);
                }
            }
        }

        if (documents.Count < 1) {
            log.Debug("No documents to index");
            return;
        }

        foreach (DocumentInfo info in documents) {
            idx.Reader.Delete(new Term("guid", info.guid));
        }

        idx.CloseReader();

        int count = 0;

        IStore abStore = new NmapStore(user, "addressbook", client);
        IStore calStore = new NmapStore(user, "calendar", client);
        IStore mailStore = new NmapStore(user, "mail", client);

        foreach (DocumentInfo info in documents) {
            try
            {
            switch(info.type) {
            case DocumentType.Addressbook :
                ab.IndexDocument(idx, abStore, info.guid);
                break;
            case DocumentType.AddressbookFolder :
                folder.IndexDocument(idx, abStore, info.guid);
                break;
            case DocumentType.Calendar :
                cal.IndexDocument(idx, calStore, info.guid);
                break;
            case DocumentType.CalendarFolder :
                folder.IndexDocument(idx, calStore, info.guid);
                break;
            case DocumentType.Mail :
                mail.IndexDocument(idx, mailStore, info.guid);
                break;
            case DocumentType.MailFolder :
                folder.IndexDocument(idx, mailStore, info.guid);
                break;
            default :
                break;
            }
            client.Document.DINSET(info.guid, 1);
            count++;

            } catch (Exception e) {
                log.ErrorFormat("Exception indexing document {0}: {1}", info.guid, e);
            }
            
            /* Batch in chunks of 20 documents for now */
            if ((args.priority != HulaThreadPriority.Interactive) && (count >= 20))
            {
                if (documents.Count > count) {
                    pool.QueueWorkToDo(new WaitCallback(IndexNMAPUser), args,
                                       args.priority);
                }
                break;
            }
        }
        abStore.Dispose();
        calStore.Dispose();
        mailStore.Dispose();
    }

    public IList SearchUser (string user, string query) {
        LuceneIndex idx = GetIndex (user);

        log.DebugFormat("Searching: {0} ({1})", query, user);

        Query q = idx.QueryParser.Parse(query);

        IList ret = new ArrayList();

        try {
            using(idx) {
                Hits hits = idx.Searcher.Search (q);

                for (int i=0; i<hits.Length(); i++) {
                    Document doc = hits.Doc(i);

                    IDictionary retdoc = new Hashtable();
                    retdoc.Add("guid", doc.Get("guid"));
                    retdoc.Add("summary", doc.Get("summary"));
                    ret.Add(retdoc);
                }
            }
        } catch (Exception e) {
            log.ErrorFormat("Exception indexing user {0}: {1}", user, e);
        }

        return ret;
    }
}
}
