/****************************************************************************
 *
 * Copyright (c) 2001-2002 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/
using System;
using System.Runtime.InteropServices;
using System.Text;

using MDBHandle = System.IntPtr;

namespace Hula.Mdb {

	public class Mdb : IDisposable {
	
		public const string HULAMDB_LIB = "hulamdb";
		public const string HULAMSGAPI_LIB = "hulamsgapi";
		public const string HULA_MODULE = "Hula";	
	
		public static readonly IntPtr NULL = System.IntPtr.Zero;
	
		private bool isInitialized = false;
		private MDBHandle handle = NULL;
		
		private int apiVersion = -1;
		private int apiCompatibleVersion = -1;
		private string apiDescription;
		private string hostDn;
		private string hostTree;

		/**
		 * State of internal MDB driver.
		 */
		public bool IsInitialized {
			get {
				return isInitialized;
			}
		}
		
		public bool IsAuthenticated {
			get {
				return handle != NULL;
			}
		}
		
		/**
		 * MDB API version number.
		 */
		public int APIVersion {
			get {
				if (apiVersion < 0 && IsInitialized) {
					apiVersion = MDBGetAPIVersion(false, null, NULL);
				}
				return apiVersion;
			}
		}
		
		/**
		 * MDB API that this implementation is compatible with.
		 */
		public int APICompatibleVersion {
			get {
				if (apiCompatibleVersion < 0 && IsInitialized) {
					apiCompatibleVersion = MDBGetAPIVersion(true, null, NULL);
				}
				return apiCompatibleVersion;
			}
		}
		
		/**
		 * MDB API description.
		 *
		 * BROKEN: This doesn't work right now because I'm not sure how to get
		 * a hold of the MDBHandle context that contains the data.
		 */
		public string APIDescription {
			get {
				/* Description field is empty without the proper context */
				MDBHandle context = NULL;
				
				if (apiDescription == null && IsInitialized) {
					StringBuilder sb = new StringBuilder(256);
					MDBGetAPIVersion(false, sb, context);
					apiDescription = sb.ToString();
				}
				return apiDescription;
			}
		}
		
		public string HostDn {
			get {
				if (hostDn == null && IsInitialized) {
					StringBuilder sb = new StringBuilder(256);
					MDBGetServerInfo(sb, null, NULL);
					hostDn = sb.ToString();
				}
				return hostDn;
			}
		}
		
		public string HostTree {
			get {
				if (hostTree == null && IsInitialized) {
					StringBuilder sb = new StringBuilder(256);
					MDBGetServerInfo(null, sb, NULL);
					hostTree = sb.ToString();
				}
				return hostTree;
			}
		}

		public Mdb() {
			Init();
		}
		
		public Mdb(string username, string password) {
			Init();
			Authenticate(username, password);
		}
		
		~Mdb() {
			Cleanup();
		}
		
		public void Dispose() {
			Cleanup();
			System.GC.SuppressFinalize(this);
		}
		
		/**
		 * MDB must be initialized before it can be used.
		 */
		private void Init() {
			MsgDirectoryHandle(); /* We call this to force the msgapi library 
									 to load. For some reason MDBInit fails w/o
									 this on Linux under Mono. */
			isInitialized = MDBInit();
		}
		
		/**
		 * Releases resources held by the object.
		 */
		private void Cleanup() {
			Release();
			MDBShutdown(); /* I don't think this acutally does anything yet */
		}

		public bool Authenticate(string username, string password) {
			handle = MDBAuthenticate(HULA_MODULE, username, password);
			if (handle == NULL) {
				return false;
			}
			return true;
		}
		
		/**
		 * Release authentication resources.
		 */
		public bool Release() {
			if (!MDBRelease(handle)) {
				return false;
			}
			return true;
		}

		public bool VerifyPassword(string dn, string password) {
			return MdbObject.VerifyPassword(handle, dn, password);
		}

		public MdbObject CreateObject(string dn, string Class) {
			MdbObject obj = null;
			
			switch (Class) {
				case MdbObject.C_USER:
					if (dn.EndsWith("\\"))
						dn.Remove(dn.LastIndexOf("\\"), 1);
					string surname = dn.Substring(dn.LastIndexOf("\\") + 1);
					obj = MdbUserObject.CreateUserObject(handle, dn, surname);
					break;
				default:
					obj = MdbObject.CreateObject(handle, dn, Class);
					break;
			}
			
			return obj;
		}
		
		public MdbUserObject CreateUserObject(string dn, string surname) {
			return MdbUserObject.CreateUserObject(handle, dn, surname);
		}
		
		public MdbObject GetObject(string dn) {
			return MdbObject.GetObject(handle, dn);
		}
		
		public string[] getObjectListEx(string context) 
		{	
			return MdbObject.GetObjectListEx(handle, context);
		}

		public string[] getObjectList(string context) {
			
			return MdbObject.GetObjectList(handle, context);
		}
		
		public bool DeleteObject(MdbObject obj) {
			return MdbObject.DeleteObject(obj);
		}
		
		public bool DeleteObject(string dn) {
			return MdbObject.DeleteObject(MdbObject.GetObject(handle, dn));
		}
		
		/**
		 * False might indicate failure as well as non-existance.
		 */
		public bool ObjectExists(string dn) {
			return MdbObject.ObjectExists(handle, dn);
		}
		
		public override string ToString() {
			string outstr = "";
			
			outstr += "MDB Library Info\n";
			outstr += "----------------\n";
			outstr += "API Version: " + APIVersion + "\n";
			outstr += "API Compatible Version: " + APICompatibleVersion + "\n";
			outstr += "API Description: " + APIDescription + "\n";
			outstr += "Host DN: " + HostDn + "\n";
			outstr += "Host Tree: " + HostTree + "\n";
			
			return outstr;
		}

		/* Native MDB functions */
		[DllImport(HULAMDB_LIB)]
		private static extern bool MDBInit();
		
		[DllImport(HULAMDB_LIB)]
		private static extern bool MDBShutdown();
		
		[DllImport(HULAMDB_LIB)]
		private static extern int MDBGetAPIVersion(
				bool wantCompatibleVersion, StringBuilder description, 
				MDBHandle context);
		
		[DllImport(HULAMDB_LIB)]
		private static extern bool MDBGetServerInfo(
				StringBuilder hostDN, StringBuilder hostTree, 
				IntPtr valueStruct);
				
		[DllImport(HULAMDB_LIB)]
		private static extern MDBHandle MDBAuthenticate(
				string module, string obj, string password);
		
		[DllImport(HULAMDB_LIB)]
		private static extern bool MDBRelease(MDBHandle handle);
		
		/* We have to call some msgapi function to force it to load for MDB 
		   This might be a bug.  This seems to happen only on Linux when 
		   running under Mono. */
                /* I had to do a similar thing to get CalCmd.cs working, so if
                   someone figures out how to fix this, look there too :-) CADV */
		[DllImport(HULAMSGAPI_LIB)]
		private static extern MDBHandle MsgDirectoryHandle();
	}
}
