/****************************************************************************
 *
 * Copyright (c) 2001-2002 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/
using System;
using System.Runtime.InteropServices;
using System.Text;
using System.Collections;

using MDBHandle = System.IntPtr;

namespace Hula.Mdb {

	public struct MDBValueStruct {
		public IntPtr Value;
		public int Used;
		public int ErrNo;
		public IntPtr Interface;
	}

	public class MdbObject : IDisposable {
		
		public static readonly IntPtr NULL = System.IntPtr.Zero;
		
		public const string C_USER = "User";
		public const string A_SURNAME = "Surname";
		public const string A_GIVEN_NAME = "Given Name";
		public const string A_FULL_NAME = "Full Name";
	
		protected IntPtr v = NULL;
		protected string dn;
		protected string _class;
		
		public string Dn {
			get {
				return dn;
			}
		}
		
		public string Class {
			get {
				return _class;
			}
		}

		public MdbObject(string dn, string _class) {
			this.dn = dn;
			this._class = _class;
		}
		
		~MdbObject() {
			Cleanup();
		}
		
		public void Dispose() {
			Cleanup();
			System.GC.SuppressFinalize(this);
		}
		
		private void Cleanup() {
			MDBDestroyValueStruct(v);
		}
		
  		public static bool VerifyPassword(MDBHandle handle, string dn, string password) {
			IntPtr v = MDBCreateValueStruct(handle, null);
			if (v == NULL) {
				return false;
 			}

			try {
				return MDBVerifyPassword(dn, password, v);
			} finally {
				MDBDestroyValueStruct(v);
			}
		}

		public static MdbObject CreateObject(MDBHandle handle, string dn, string Class) {
			MdbObject obj = new MdbObject(dn, Class);
		
			obj.v = MDBCreateValueStruct(handle, obj.dn);
			if (obj.v == NULL) {
				return null;
			}
			
			IntPtr attribute = MDBShareContext(obj.v);
			if (attribute == NULL) {
				MDBDestroyValueStruct(obj.v);
				return null;
			}
			
			IntPtr data = MDBShareContext(obj.v);
			if (data == NULL) {
				MDBDestroyValueStruct(attribute);
				MDBDestroyValueStruct(obj.v);
				return null;
			}
			
			if (!MDBCreateObject(obj.dn, obj.Class, attribute, data, obj.v)) {
				MDBDestroyValueStruct(data);
				MDBDestroyValueStruct(attribute);
				MDBDestroyValueStruct(obj.v);
				return null;
			}
			
			MDBDestroyValueStruct(data);
			MDBDestroyValueStruct(attribute);
			
			return obj;
		}
		
		public static MdbObject GetObject(MDBHandle handle, string dn) {
			MdbObject obj = null;
			
			IntPtr v = MDBCreateValueStruct(handle, dn);
			if (v == NULL) {
				return null;
			}
			
			StringBuilder classSB = new StringBuilder(256);
			StringBuilder dnSB = new StringBuilder(256);
			
			if (MDBGetObjectDetails(dn, classSB, null, dnSB, v)) {
				obj = new MdbObject(dnSB.ToString(), classSB.ToString());
				obj.v = v;
			}
			
			return obj;
		}
		
		public static string[] GetObjectListEx(MDBHandle handle, string container, string Class, string filter) {
			ArrayList resultList = new ArrayList();
			
			IntPtr v = MDBCreateValueStruct(handle, container);
			if (v == NULL) {
				return null;
			}

			IntPtr e = MDBCreateEnumStruct( v );
			if (e == NULL) {
				return null;
			}
		
			string x = MDBEnumerateObjectsEx(container, Class, filter, (uint)1, e, v);
			while (x != null) {
				resultList.Add(x);		
				x = MDBEnumerateObjectsEx(container, Class, filter, (uint)1, e, v);
			}
			
			MDBDestroyEnumStruct(e,v);
			MDBDestroyValueStruct(v);

			return(string[]) resultList.ToArray(typeof(string));
		}

		public static string[] GetObjectListEx(MDBHandle handle, string container) {
			
			return GetObjectListEx(handle, container,null, null);
		}

		public static string[] GetObjectList(MDBHandle handle, string container, string Class, string filter) {
			IntPtr v = MDBCreateValueStruct(handle, container);
			if (v == NULL) {
				return null;
			}
			
			int count = MDBEnumerateObjects( container, Class, filter, v);

			MDBValueStruct mdbvs = (MDBValueStruct) Marshal.PtrToStructure(v, typeof(MDBValueStruct));
			
			if (mdbvs.Value == NULL) {
				return new string[]{};
			}
			
			string[] values = new string[count];
			for (int i = 0; i < count; i++) {
				IntPtr str = Marshal.ReadIntPtr(mdbvs.Value, i * IntPtr.Size);
				values[i] = Marshal.PtrToStringAnsi(str);
			}
			
			MDBDestroyValueStruct(v);
			return values;
		}

		public static string[] GetObjectList(MDBHandle handle, string container) {
			
			return GetObjectList(handle, container, null, null);
		}

		public static bool DeleteObject(MdbObject obj) {
			if (!MDBDeleteObject(obj.dn, true, obj.v))
				return false;

			//obj.Dispose();
			return true;
		}
		
		/**
		 * False might indicate failure as well as non-existance.
		 */
		public static bool ObjectExists(MDBHandle handle, string dn) {
			IntPtr v = MDBCreateValueStruct(handle, dn);
			if (v == NULL) {
				return false;
			}
			
			bool result = MDBIsObject(dn, v);
			MDBDestroyValueStruct(v);

			return result;
		}
		
		public bool AddAttribute(string attribute, string value) {
			if (v == NULL)
				return false;
			
			if (!MDBAdd(dn, attribute, value, v))
				return false;
			
			return true;
		}
		
		public bool RemoveAttribute(string attribute, string value) {
			if (v == NULL)
				return false;
			
			if (!MDBRemove(dn, attribute, value, v))
				return false;
			
			return true;
		}
		
		public string[] GetAttribute(string attribute) {
			IntPtr tmpvs = MDBShareContext(v);
			
			int count = MDBRead(dn, attribute, tmpvs);
			MDBValueStruct mdbvs = (MDBValueStruct) Marshal.PtrToStructure(tmpvs, typeof(MDBValueStruct));
			
			if (mdbvs.Value == NULL) {
				return new string[]{};
			}
			
			string[] values = new string[count];
			for (int i = 0; i < count; i++) {
				IntPtr str = Marshal.ReadIntPtr(mdbvs.Value, i * IntPtr.Size);
				values[i] = Marshal.PtrToStringAnsi(str);
			}
			
			MDBDestroyValueStruct(tmpvs);
			return values;
		}

		public string[] GetAttributeEx(string attribute) {
			
			IntPtr tmpvs = MDBShareContext(v);
			IntPtr e = MDBCreateEnumStruct( tmpvs );
			if (e == NULL) {
				return null;
			}

			ArrayList resultList = new ArrayList();

			string x = MDBReadEx(dn, attribute, e, tmpvs);
			while (x != null) {
				resultList.Add(x);		
				x = MDBReadEx(dn, attribute, e, tmpvs);
			}
			
			MDBDestroyEnumStruct(e,tmpvs);
			MDBDestroyValueStruct(tmpvs);

			return(string[]) resultList.ToArray(typeof(string));
		}
		
		public bool SetAttribute(string attribute, string[] values) {
			foreach (string value in values) {
				if (!MDBAddValue(value, v))
					return false;
			}
			
			if (!MDBWrite(dn, attribute, v))
				return false;
			
			return true;
		}
		
		public bool SetAttribute(string attribute, string value) {
			return SetAttribute(attribute, new string[1]{value});
		}

		
		/**
		 * Changes the password.  If 'oldPassword' is null uses 
		 * MDBChangePasswordEx.
		 *
		 * NOTE*** Make sure you have MemMgr initialized before calling this
		 * function if you are using the MDB File driver.
		 */
		public bool ChangePassword(string oldPassword, string newPassword) {
			if (v == NULL)
				return false;
			
			if (oldPassword == null) {
				if (!MDBChangePasswordEx(dn, "", newPassword, v))
					return false;
			} else {
				if (!MDBChangePassword(dn, oldPassword, newPassword, v))
					return false;
			}
			
			return true;
		}
		
		public override string ToString() {
			string outstr = "";
			
			outstr += "MDB Object Info\n";
			outstr += "---------------\n";
			outstr += "DN: " + Dn + "\n";
			outstr += "Class: " + Class + "\n";
			
			return outstr;
		}
		
		/* Native MDB functions */
		/* MDBValueStruct *	MDBCreateValueStruct(
								MDBHandle	Handle,
								const unsigned char	*Context);
		*/
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern IntPtr MDBCreateValueStruct(
				MDBHandle handle, string context);
		
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBDestroyValueStruct(IntPtr valueStruct);
				
		/* MDBValueStruct *	MDBShareContext(
								MDBValueStruct	*Vorg); */
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern IntPtr MDBShareContext(IntPtr valueStruct);
				
		/* BOOL	MDBSetValueStructContext(
					const unsigned char	*Context,
					MDBValueStruct	*V); */
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBSetValueStructContext(string context, IntPtr valueStruct);

		/* 
		 *MDBEnumStruct	*MDBCreateEnumStruct( MDBValueStruct	*V ); 
		 */
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern IntPtr MDBCreateEnumStruct( IntPtr valueStruct );
		 
		/* 
		 * BOOL	MDBDestroyEnumStruct(
					MDBEnumStruct	*E,
					MDBValueStruct	*V );
		 */
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBDestroyEnumStruct(IntPtr enumStruct, IntPtr valueStruct);

		/* BOOL	MDBAddValue(
					const unsigned char	*Value,
					MDBValueStruct	*V ); */
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBAddValue(string value, IntPtr valueStruct);
				
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern int MDBRead(string obj, string attribute, IntPtr valueStruct);

		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern string MDBReadEx(string obj, string attribute, IntPtr enumStruct, IntPtr valueStruct);
		
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern int MDBReadDN(string obj, string attribute, IntPtr valueStruct);
				
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBWrite(string obj, string attribute, IntPtr valueStruct);
				
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBAdd(
				string obj, string attribute, string val, IntPtr valueStruct);
		
		/* This was a macro in mdb.h */
		protected static bool MDBAddStringAttribute(
				string attribute, string value, IntPtr attributes, IntPtr data) {
			string temp = "S" + attribute;
			
			if (!MDBAddValue(temp, attributes))
				return false;
			if (!MDBAddValue(value, data))
				return false;
				
			return true;
		}
		
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBRemove(
				string obj, string attribute, string val, IntPtr valueStruct);	
			
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBCreateObject(
				string obj, string Class, IntPtr attribute, 
				IntPtr data, IntPtr v);
		
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBDeleteObject(
				string obj, bool recursive, IntPtr valueStruct);
				
		/*BOOL	MDBGetObjectDetails(
		 			const unsigned char	*Object,
					unsigned char	*Type,
					unsigned char	*RDN,
					unsigned char	*DN,
					MDBValueStruct	*V );*/

		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBGetObjectDetails(
				string dn, StringBuilder type, StringBuilder rdn, 
				StringBuilder adn, IntPtr valueStruct);

		/*long	MDBEnumerateObjects(
					const unsigned char	*Container,
					const unsigned char	*Type,
					const unsigned char	*Pattern,
					MDBValueStruct	*V );*/

		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern int MDBEnumerateObjects(
			string dn, string type, string pattern, IntPtr valueStruct);
				
		/* online doc
			* const unsigned char*	MDBEnumerateObjectsEx(
					const unsigned char	*Container,
					const unsigned char	*Type,
					const unsigned char	*Pattern,
					unsigned int Flags,			** set to true/1 if you want the result set to include containers ??
					MDBEnumStruct	*E,
					MDBValueStruct	*V
					);
			the mdbfile driver doesn't seem to respect the flag parameter
			from looking at the source, the eDir driver should filter out(0) or include(1) containers
		*/
		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern string MDBEnumerateObjectsEx(
			string dn, string type, string pattern, uint flags, IntPtr enumStruct, IntPtr valueStruct);

		/* const unsigned char	*MDBEnumerateAttributesEx(
								   const unsigned char	*Object,
								   MDBEnumStruct	*E,
								   MDBValueStruct	*V); */

		/* this function is not implemented in the mdb file driver yet (9/2005)
		 * [DllImport(Mdb.HULAMDB_LIB)]
		protected static extern string MDBEnumerateAttributesEx(
			string Object, IntPtr enumStruct, IntPtr valueStruct);*/

		[DllImport(Mdb.HULAMDB_LIB)]
		protected static extern bool MDBIsObject(string obj, IntPtr valueStruct);
		
		[DllImport(Mdb.HULAMDB_LIB)]
		private static extern bool MDBVerifyPassword(
				string obj, string password, IntPtr valueStruct);

		[DllImport(Mdb.HULAMDB_LIB)]
		private static extern bool MDBChangePassword(
				string obj, string oldPassword, string newPassword, IntPtr valueStruct);
		
		[DllImport(Mdb.HULAMDB_LIB)]
		private static extern bool MDBChangePasswordEx(
				string obj, string oldPassword, string newPassword, IntPtr valueStruct);
	}
}
