/****************************************************************************
 *
 * Copyright (c) 2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/

using System;
using System.IO;
using System.Text;

using log4net;

namespace Hula.Nmap
{
public class NmapStreamReader : TextReader
{
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    StreamReader reader = null;
    int len = 0;
    int pos = 0;
    bool finished = false;

    public NmapStreamReader(StreamReader reader, int len)
    {
        this.reader = reader;
        this.len = len;
        this.pos = 0;
        this.finished = false;

        log.DebugFormat("initialized, expecting to read {0} bytes", len);

	MaybeFinish();
    }

    private bool AtEnd ()
    {
        if (pos >= len)
            return true;

        return false;
    }

    private int BytesLeft ()
    {
        return len - pos;
    }

    private void MaybeFinish ()
    {
        if (AtEnd () && !finished)
	{
	    string response = reader.ReadLine();
            log.DebugFormat("RECV: {0}", response);
            finished = true;
        }
    }

    public override void Close ()
    {
        ReadToEnd();
    }

    public override int Peek ()    {
        return reader.Peek ();
    }

    public override int Read ()
    {
        if (AtEnd()) {
            return -1;
	}

        int ret = reader.Read ();
        pos++;

        MaybeFinish ();
        return ret;
    }

    public override int Read (char[] buf, int index, int count)
    {
        if (AtEnd())
        {
            return -1;
        }

        int left = BytesLeft();

        if (count > left)
            count = left;
     
        int bytes = reader.Read (buf, index, count);
        pos = pos + bytes;

        MaybeFinish ();
        return bytes;
    }

    public override int ReadBlock (char[] buf, int index, int count)
    {
        if (AtEnd())
            return -1;

        int left = BytesLeft();

        if (count > left)
            count = left;
        
        int bytes = reader.ReadBlock (buf, index, count);
        pos = pos + bytes;

        MaybeFinish ();
        return bytes;
    }

    public override string ReadLine ()
    {
        if (AtEnd())
            return null;

	bool foundCR = false;
	StringBuilder text = new StringBuilder ();
	
	while (true) {
	    int c = Read ();
	    
	    if (c == -1) {				// end of stream
		if (text.Length == 0)
		    return null;
		
		if (foundCR)
		    text.Length--;
		
		break;
	    }
	    
	    if (c == '\n') {			// newline
		if ((text.Length > 0) && (text [text.Length - 1] == '\r'))
		    text.Length--;
		
		foundCR = false;
		break;
	    } else if (foundCR) {
		pos--;
		text.Length--;
		break;
	    }
	    
	    if (c == '\r')
		foundCR = true;
	    
	    
	    text.Append ((char) c);
	}
	
        MaybeFinish ();

	return text.ToString ();
    }

    public override string ReadToEnd ()
    {
        if (AtEnd())
        {
            return null;
        }

        char[] buf = new char[BytesLeft()];
        int bytes = ReadBlock (buf, 0, BytesLeft());
        pos = pos + bytes;

        MaybeFinish ();
        return new string(buf);
    }
}
}
