using System;
using System.IO;
using System.Collections;

namespace Hula.Store {

// <summary>
// Indicates that an operation could not succeed because the etag was invalid.
// </summary>
public class ETagException : System.Exception {
    public ETagException () : base(){
    }

    public ETagException (string description, Exception e) : base(description, e){
    }

    public ETagException (Exception e) : base("Conflict", e){
    }
} 

// <summary>
// Manages a user store.
// </summary>
public interface IStore : IDisposable {
        
    string Owner { get; }
    string Name { get; }

    //<summary>
    // Add a stream to the store.  If the item needs to be
    // indexed, an IndexStream event will be emitted.
    //</summary>
    void Add (string folder, 
	      IDictionary properties, 
	      TextReader reader,
	      string mimeType,
	      out string guid,
	      out string etag);

    //<summary>
    // Retrieves a stream from the store.
    //</summary>
    TextReader Get (string guid);

    //<summary>
    // Retrieves a stream from the store, converting it to a type listed in accept
    //</summary>
    TextReader Get (string guid, string accept, out string mimeType);

    //<summary>
    // Retrieves a stream from the store.
    //</summary>
    void Get (string guid, out IDictionary properties, out TextReader reader);

    //<summary>
    // Retrieves a stream from the store, converting it to a type listed in accept
    //</summary>
    void Get (string guid, string accept, out IDictionary properties, out TextReader reader, out string mimeType);

    string GetType (string mimeType, string accept);

    //<summary>
    // Replace a document with a new document.
    //</summary>
    void Replace (string guid, string etag, TextReader reader, string mimeType, out string newEtag);

    //<summary>
    //  Move a document to a different folder.
    //</summary>
    void Move (string guid, string etag, string newFolder);
    
    //<summary>
    // Copy a document to a folder.
    //</summary>
    //<return>
    // Returns the guid of the newly created document
    //</return>
    string Copy (string guid, string etag, string newFolder);
    
    //<summary>
    // Delete a document from the store.
    //</summary>
    void Delete (string guid, string etag);
    
    string AddFolder(string folderName);
    
    //<summary>
    // Returns the properties attached to a guid.
    //</summary>
    //<return>
    // An IDictionary containing all the properties attached to a guid
    //</return>
    IDictionary GetProperties (string guid);
    
    //<summary>
    // Returns a property's value.  If the property is a live (server-managed) property, 
    // the server will validate the value before setting it.  If the property is a dead (client-managed) property, 
    // it will be set.
    //</summary>
    string GetProperty (string guid, string key);
    
    //<summary>
    // Sets a property's value.
    //</summary>
    void SetProperty (string guid, string key, string value);
}
}
