using System;
using System.Collections;
using System.Threading;

// Use Lucene's PriorityQueue implementation
using Lucene.Net.Util;

using log4net;

namespace Hula.Util {
    public enum HulaThreadPriority {
        Interactive = 0,
        NonInteractive = 1,
        Bulk = 2
    }

    public class ThreadInfo {
        public ThreadStart ThreadStart;

        public WaitCallback WaitCallback;
        public object UserData;
        public HulaThreadPriority Priority;
        public DateTime QueueTime;

        public static ThreadInfo Zero = new ThreadInfo();
    }

    internal class HulaThread {
	private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        private static int nextId = 500;
        
        private Thread thread;
        public ThreadInfo tinfo;
        private bool running;
        private DateTime finishTime = DateTime.MinValue;
        private int id;
        
        public event EventHandler Finished;

        public int Id {
            get { return id; }
        }
        
        public DateTime FinishTime {
            get { return finishTime; }
        }

        public HulaThread () {
            id = nextId++;
            running = true;
            thread = new Thread (new ThreadStart (WorkerLoop));
            tinfo = ThreadInfo.Zero;
            thread.IsBackground = true;
            thread.Start ();
        }

        public void RunThread (ThreadInfo tinfo) {
            lock (this) {
                log.DebugFormat ("Thread {0}: got new worker", id);
                this.tinfo = tinfo;
                Monitor.Pulse (this);
            }
        }

        public void Stop () {
            lock (this) {
                log.DebugFormat ("Thread {0}: stopping", id);
                running = false;
                Monitor.Pulse (this);
            }
        }

        private void WorkerLoop () {
            while (running) {
                lock (this) {
                    if (tinfo.Equals (ThreadInfo.Zero)) {
                        log.DebugFormat ("Thread {0}: waiting on job...", id);
                        Monitor.Wait (this);
                    }

                    if (!running) {
                        log.DebugFormat ("Thread {0}: terminating", id);
                        break;
                    }

                    try {
                        log.DebugFormat ("Thread {0}: running job...", id);
                        if (tinfo.ThreadStart != null) {
                            tinfo.ThreadStart ();
                        } else {
                            tinfo.WaitCallback (tinfo.UserData);
                        }
                    } catch (Exception e) {
                        log.ErrorFormat ("Thread {0}: unhandled exception: {1}:", id, e);
                    } finally {
                        log.DebugFormat ("Thread {0}: job finished", id);
                        finishTime = DateTime.Now;

                        if (Finished != null)
                            Finished (this, null);
                    }
                }
            }
        }
    }
    
    public class HulaThreadPool {
	private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        private static TimeSpan ThreadTimeout = TimeSpan.FromMinutes (10);

        private ArrayList readyThreads = new ArrayList ();
        private ArrayList activeThreads = new ArrayList ();
        
        private IndexingQueue queue = new IndexingQueue ();

        private int max = 50;
        private int min = 3;
        
        public int MaxThreads {
            get { return max; }
            set { max = value; }
        }

        public int MinThreads {
            get { return min; }
            set { min = value; }
        }

        private int TotalThreads {
            get {
                lock (readyThreads) {
                    return readyThreads.Count + activeThreads.Count;
                }
            }
        }

        public HulaThreadPool () : this (3, 50) {
        }

        public HulaThreadPool (int min, int max) {
            this.min = min;
            this.max = max;

            for (int i = 0; i < min; i++) {
                AddThread (ThreadInfo.Zero);
            }
        }

        private void AddThread (ThreadInfo tinfo) {
            lock (readyThreads) {
                HulaThread thread = new HulaThread ();

                log.DebugFormat ("Added a new thread ({0}), total count is {1}", thread.Id, (TotalThreads + 1));
                
                thread.Finished += OnThreadFinished;

                if (!tinfo.Equals (ThreadInfo.Zero)) {
                    activeThreads.Add (thread);
                    thread.RunThread (tinfo);
                } else {
                    thread.tinfo = ThreadInfo.Zero;
                    readyThreads.Add (thread);
                }
            }
        }

        private void OnThreadFinished (object o, EventArgs args) {
            lock (readyThreads) {
                HulaThread thread = (HulaThread) o;

                queue.Finish(thread.tinfo);
                thread.tinfo = ThreadInfo.Zero;

                activeThreads.Remove (thread);
                readyThreads.Add (thread);

                if (queue.Size() > 0) {
                    TryDoWork ();
                }
            }
        }

        public void QueueWorkToDo (WaitCallback cb, object data, HulaThreadPriority priority) {
            ThreadInfo tinfo = new ThreadInfo ();
            tinfo.WaitCallback = cb;
            tinfo.UserData = data;
            tinfo.Priority = priority;
            tinfo.QueueTime = DateTime.Now;

            RunThread (tinfo);
        }

        private void RunThread (ThreadInfo tinfo) {
            lock (readyThreads) {
                log.Debug ("Queueing new job");
                queue.Put (tinfo);

                TryDoWork ();
            }
        }

        private void TryDoWork ()
        {
            lock (readyThreads) {
                CleanupThreads ();

                if (queue.Size () == 0)
                {
                    return;
                }

                // find an unused thread if possible
                if (readyThreads.Count > 0) {
                    HulaThread thread = (HulaThread) readyThreads[0];
                    readyThreads.Remove (thread);
                    activeThreads.Add (thread);
                    
                    log.DebugFormat ("Running job in inactive thread {0}", thread.Id);
                    ThreadInfo tinfo = (ThreadInfo) queue.Pop ();

                    if (tinfo != null)
                    {
                        thread.RunThread (tinfo);
                    }

                    return;
                }

                // no free threads, see if we can create a new one
                if (TotalThreads < max) {
                    ThreadInfo tinfo = (ThreadInfo) queue.Pop ();

                    if (tinfo != null)
                    {
                        AddThread (tinfo);
                    }

                    return;
                }
            }
        }

        private void CleanupThreads () {
            lock (readyThreads) {
                
                foreach (HulaThread thread in readyThreads.ToArray (typeof (HulaThread))) {
                    if (TotalThreads == min)
                        break;
                    
                    if (DateTime.Now - thread.FinishTime > ThreadTimeout || TotalThreads > max) {

                        log.DebugFormat ("Stopping thread {0}", thread.Id);
                        thread.Stop ();
                        readyThreads.Remove (thread);
                    }
                }
            }
        }
    }
}

