using System;
using System.Collections;
using System.IO;
using System.Net;
using System.Xml;
using System.Xml.XPath;
using NUnit.Framework;
using Hula.TestData;

namespace Hula.Dav {
[TestFixture]
public class TestDav {
    public static int testPort = 8080;
    public static string testUser = "test";
    public string testPassword = "test";

    public static void CleanupTestDirs() {
	// Blast the index and store out from underneath the running
	// server.
	// This is really ugly.
	DeleteDir("../src/store-" + testUser);
	DeleteDir("../src/index-" + testUser);
    }

    public static void DeleteDir(string dirname) {
	if (System.IO.Directory.Exists(dirname)) {
	    System.IO.Directory.Delete(dirname, true);
	}
    }

    [SetUp]
    public void SetUp() {
	CleanupTestDirs();
    }

    [TearDown]
    public void TearDown() {
    }

    public WebRequest GetRequest(string method, string path, string user){
	WebRequest request = WebRequest.Create("http://127.0.0.1:" + testPort.ToString() + "/dav/" + user + "/" + path);
	request.Credentials = Credentials();
	request.PreAuthenticate = true;
	request.Method = method;
	return request;
    }

    public WebRequest GetRequest(string method, string path){
	return GetRequest(method, path, testUser);
    }

    public ICredentials Credentials() {
	return new NetworkCredential(testUser, testPassword);
    }

    public void IgnoreResponse(WebRequest r) {
	r.GetResponse().Close();
    }
    public void CreateCollection(string folderName) {
	IgnoreResponse(GetRequest("MKCOL", folderName));
    }

    public void CreateCalendar(string calName) {
	string folder = testUser + "/calendar/" + calName + "/";
	CreateCollection(folder);
    }

    [Test]
    public void TestMkRoot() {
	bool success = false;
	try {
	    // Can't create a root directory
	    CreateCollection("/");
	} catch (WebException e) {
	    if (((HttpWebResponse)e.Response).StatusCode == HttpStatusCode.MethodNotAllowed) {
		success = true;
	    } else {
		Assert.Fail(e.ToString());
	    }
	}
	if (!success) {
	    Assert.Fail("No exception");
	}
    }

    [Test]
    public void TestMkStore() {
	bool success = false;
	try {
	    // Can't create a store
	    CreateCollection("/calendar/");
	} catch (WebException e) {
	    if (((HttpWebResponse)e.Response).StatusCode == HttpStatusCode.MethodNotAllowed) {
		success = true;
	    } else {
		Assert.Fail(e.ToString());
	    }
	}
	if (!success) {
	    Assert.Fail("No exception");
	}
    }

    [Test]
    public void TestMissingParent() {
	bool success = false;
	try {
	    // Can't create a with a missing parent
	    CreateCollection("/calendar/TestCal/Blah/");	
	} catch (WebException e) {
	    if (((HttpWebResponse)e.Response).StatusCode == HttpStatusCode.Conflict) {
		success = true;
	    } else {
		Assert.Fail(e.ToString());
	    }
	}
	if (!success) {
	    Assert.Fail("No exception");
	}
    }

    [Test]
    public void TestCreateCal() {
	CreateCollection("/calendar/TestCal/");
    }

    [Test]
    public void TestSubCal() {
	// Calendars must be immediate children of /calendar/
	CreateCollection("/calendar/TestCal/");
	bool success = false;
	try {
	    CreateCollection("/calendar/TestCal/Blah/");
	} catch (WebException e) {
	    if (((HttpWebResponse)e.Response).StatusCode == HttpStatusCode.MethodNotAllowed) {
		success = true;
	    } else {
		Assert.Fail(e.ToString());
	    }
	}
	if (!success) {
	    Assert.Fail("No exception");
	}

    }

    public WebResponse Put(string path, string ifNoneMatch, string ifMatch, string contentType, string data) {
	WebRequest request = GetRequest("PUT", path);
	if (ifNoneMatch != null) {
	    request.Headers["If-None-Match"] = ifNoneMatch;
	}
	if (ifMatch != null) {
	    request.Headers["If-Match"] = ifMatch;
	}

	request.ContentType = contentType;

	using(StreamWriter writer = new StreamWriter(request.GetRequestStream())) {
	    writer.Write(data);
	}

	return request.GetResponse();
    }

    public WebResponse Get(string path, string accept) {
	WebRequest request = GetRequest("GET", path);
	if (accept != null) {
	    ((HttpWebRequest)request).Accept = accept;
	}

	return request.GetResponse();
    }

    public WebResponse Report(string path, string data) {
	WebRequest request = GetRequest("REPORT", path);
	request.ContentLength = data.Length;
	request.ContentType = "text/xml";

	using(StreamWriter writer = new StreamWriter(request.GetRequestStream())) {
	    writer.Write(data);
	    writer.Flush();
	}

	return request.GetResponse();
    } 

    [Test]
    public void TestIfNoneMatch() {
	    CreateCollection("/calendar/TestCal/");
	    
	    Put("/calendar/TestCal/meeting1_1.ics", "*", null, "text/calendar", Data.GetString("meeting1_1.ics")).Close();
	    
	bool success = false;
	try {
	    // This should fail
	    Put("/calendar/TestCal/meeting1_1.ics", "*", null, "text/calendar", Data.GetString("meeting1_1.ics")).Close();
	} catch (WebException e) {
	    if (((HttpWebResponse)e.Response).StatusCode == HttpStatusCode.Conflict) {
		success = true;
	    } else {
		Assert.Fail(e.ToString());
	    }
	}
	if (!success) {
	    Assert.Fail("No exception");
	}
    }

    [Test]
    public void TestIfMatchFail() {
	CreateCollection("/calendar/TestCal/");
	
	Put("/calendar/TestCal/meeting1_1.ics", "*", null, "text/calendar", Data.GetString("meeting1_1.ics")).Close();
	    
	bool success = false;
	try {
	    // This should fail
	    Put("/calendar/TestCal/meeting1_1.ics", null, "bad-etag", "text/calendar", Data.GetString("meeting1_1.ics")).Close();
	} catch (WebException e) {
	    if (((HttpWebResponse)e.Response).StatusCode == HttpStatusCode.Conflict) {
		success = true;
	    } else {
		Assert.Fail(e.ToString());
	    }
	}
	if (!success) {
	    Assert.Fail("No exception");
	}

    }

    [Test]
    public void TestIfMatchSucceed() {
	CreateCollection("/calendar/TestCal/");
	
	WebResponse response = Put("/calendar/TestCal/meeting1_1.ics",
				   "*", null, 
				   "text/calendar", Data.GetString("meeting1_1.ics"));
	string etag = response.Headers["ETag"];
	response.Close();
	Put("/calendar/TestCal/meeting1_1.ics", 
	    null, etag,
	    "text/calendar", Data.GetString("meeting1_1.ics")).Close();
    }

    private XmlNamespaceManager CreateNamespaceManager(XmlNameTable table) {
	XmlNamespaceManager m = new XmlNamespaceManager(table);
	m.AddNamespace("D","DAV:");
	return m;
    }

    private void CheckHRefs(TextReader reader, IEnumerable hrefNames) {
	Hashtable hrefs = new Hashtable();
	foreach (string href in hrefNames) {
	    string fullhref = "http://127.0.0.1:8080/dav/" + testUser + href;
	    hrefs[fullhref] = true;
	}

	XmlDocument doc = new XmlDocument();
	doc.Load(reader);

	XmlNamespaceManager context = CreateNamespaceManager(doc.NameTable);
	XmlNodeList nodes = doc.SelectNodes("//D:response/D:href", context);

	foreach (XmlNode node in nodes) {
	    if (hrefs.Contains(node.InnerText)) {
		hrefs.Remove(node.InnerText);
	    } else {
		Assert.Fail("unexpected href: {0}", node.InnerText);
	    }
	}
	
	if (hrefs.Count != 0) {
	    Assert.Fail("missing results");
	}
    }


    private  void PutCalItem(string itemName) {
	WebResponse response = Put("/calendar/TestCal/" + itemName,
				   "*", null, 
				   "text/calendar", Data.GetString(itemName));
	response.Close();
    }

    private  void PutAddressbookItem(string itemName) {
	WebResponse response = Put("/addressbook/TestContacts/" + itemName,
				   "*", null, 
				   "text/x-vcard", Data.GetString(itemName));
	response.Close();
    }

    private void SetupCalData() {
	CreateCollection("/calendar/TestCal/");

	PutCalItem("meeting1_6.ics");
	PutCalItem("meeting2_4.ics");
	PutCalItem("meeting3_1.ics");
	PutCalItem("meeting4_1.ics");
	PutCalItem("meeting5_1.ics");
    }

    private void SetupAddressbookData() {
	CreateCollection("/addressbook/TestContacts/");

	PutAddressbookItem("card1.vcf");
	PutAddressbookItem("card2.vcf");
    }

    private void QueryCal(string filename, 
			      IEnumerable filenames) {
	ArrayList hrefs = new ArrayList();
	foreach (string file in filenames) {
	    hrefs.Add("/calendar/TestCal/" + file);
	}
	WebResponse response = Report("/calendar/TestCal/",
				      Data.GetString(filename));
	using (TextReader reader = new StreamReader(response.GetResponseStream())) {	    
	    CheckHRefs(reader, hrefs);
	}
    }

    private void QueryAddressbook(string filename, 
				  IEnumerable filenames) {
	ArrayList hrefs = new ArrayList();
	foreach (string file in filenames) {
	    hrefs.Add("/addressbook/TestContacts/" + file);
	}
	WebResponse response = Report("/addressbook/TestContacts/",
				      Data.GetString(filename));
	using (TextReader reader = new StreamReader(response.GetResponseStream())) {	    
	    CheckHRefs(reader, hrefs);
	}
    }

    private void CheckIsICal(WebResponse r) {
	using(StreamReader reader = new StreamReader(r.GetResponseStream())) {
	    Assert.AreEqual("text/calendar", r.ContentType);
	    Assert.AreEqual("BEGIN:VCALENDAR", reader.ReadLine());
	}
    }

    private void CheckIsXml(WebResponse r) {
	Assert.AreEqual("application/calendar+xml", r.ContentType);
	XmlDocument d = new XmlDocument();
	d.Load(r.GetResponseStream());
    }

    [Test]
    public void TestAccept() {
	SetupCalData();

	WebResponse r = Get("/calendar/TestCal/meeting2_4.ics", null);
	CheckIsICal(r);

	r = Get("/calendar/TestCal/meeting2_4.ics", "text/calendar");
	CheckIsICal(r);

	r = Get("/calendar/TestCal/meeting2_4.ics", "application/calendar+xml, text/calendar");
	CheckIsICal(r);

	r = Get("/calendar/TestCal/meeting2_4.ics", "application/calendar+xml, text/calendar, text/html");
	CheckIsICal(r);

	r = Get("/calendar/TestCal/meeting2_4.ics", "application/calendar+xml");
	CheckIsXml(r);

	r = Get("/calendar/TestCal/meeting2_4.ics", "application/calendar+xml, text/html");
	CheckIsXml(r);

	r = Get("/calendar/TestCal/meeting2_4.ics", "application/calendar+xml, text/html");
	CheckIsXml(r);
    }

    [Test]
    public void TestReportAll() {
	SetupCalData();
	SetupAddressbookData();
	
	QueryCal("query1.xml",
		  new string[] {
	            "meeting1_6.ics",
			"meeting2_4.ics",
			"meeting3_1.ics",
			"meeting4_1.ics",
			"meeting5_1.ics" });
    }

    [Test]
    public void TestReportEvent() {
	SetupCalData();
	
	// Add a TODO 
	PutCalItem("todo1.ics");

	QueryCal("query3_1.xml",
		  new string[] {
	            "todo1.ics" });
	
	QueryCal("query3.xml",
		  new string[] {
	            "meeting1_6.ics",
			"meeting2_4.ics",
			"meeting3_1.ics",
			"meeting4_1.ics",
			"meeting5_1.ics" });
    }

    [Test]
    public void TestTimeQuery() {
	SetupCalData();
	QueryCal("query6.xml",
		     new string[] {
	            "meeting1_6.ics",
			"meeting2_4.ics",
			"meeting3_1.ics",
			"meeting4_1.ics"});

	QueryCal("query6_1.xml",
		     new string[] {
	             "meeting5_1.ics"});
    }

    [Test]
    public void TestRecurringTimeQuery() {
	SetupCalData();
	QueryCal("query5.xml",
		     new string[] { "meeting2_4.ics" });
    }

    [Test]
    public void TestPropTextMatchQuery() {
	SetupCalData();
	QueryCal("query10.xml",
		     new string[] { "meeting5_1.ics" });
    }

    [Test]
    public void TestPropTextMatchWrongCaseQuery() {
	SetupCalData();
	QueryCal("query10.1.xml",
		 new string[] { });
    }

    [Test]
    public void TestSimpleParamQuery() {
	SetupCalData();
	QueryCal("query12.xml",
		 new string[] { "meeting2_4.ics" });
    }

    [Test]
    public void TestVCardPut() {
	CreateCollection("/addressbook/TestContacts/");

	Put("/addressbook/TestContacts/card1.vcf",
	    "*", null,
	    "text/x-vcard", Data.GetString("card1.vcf")).Close();
    }

    [Test]
    public void TestVCardQuery() {
	SetupAddressbookData();
	QueryAddressbook("lucene-query1.xml",
			 new string[] { "card1.vcf" });
	QueryAddressbook("lucene-query2.xml",
			 new string[] { "card2.vcf" });
	QueryAddressbook("lucene-query3.xml",
			 new string[] { "card1.vcf", "card2.vcf" });
    }
}
}
